# -------------------------------------------------------------------------
#
#        D E S T I N A T I O N   S Y S T E M
#
# -------------------------------------------------------------------------
import abc
import os
import pathlib
import shutil
from abc import ABC, abstractmethod


class DeliveryContextIF(ABC):
    @abstractmethod
    def create_tempfile(self, prefix: str, suffix: str) -> pathlib.Path:
        pass

    @abstractmethod
    def token(self) -> str:
        pass


class Destination(abc.ABC):
    """
    Destinations are locations that files can be copied into. They might not
    always be on a local disk; FTP or Globus could also be destinations.

    The destination API is very simply, consisting just of adding files.
    """

    def __init__(self, context: DeliveryContextIF, path: pathlib.Path):
        self.context = context
        self.path = path

    @abc.abstractmethod
    def add_file(self, file: pathlib.Path, relative_path: str):
        pass

    @abc.abstractmethod
    def add_directory(self, directory: pathlib.Path, relative_path: str):
        pass

    @abc.abstractmethod
    # For tar; may not be necessary
    def close(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # ensure that if we are used as a context manager ('with' statement)
        # that the destinations do get properly closed
        self.close()


class DestinationDecorator(Destination):
    def __init__(self, underlying: Destination):
        self.underlying = underlying
        super().__init__(underlying.context, underlying.path)

    def add_file(self, file: pathlib.Path, relative_path: str):
        self.underlying.add_file(file, relative_path)

    def add_directory(self, directory: pathlib.Path, relative_path: str):
        self.underlying.add_directory(directory, relative_path)

    def close(self):
        self.underlying.close()


class TarDecorator(DestinationDecorator):
    """
    This decorator creates a local tar archive. Calls to add_file
    are intercepted and instead the file contents are added to the
    tar archive. When close() is called, we finalize the tar file
    and place it in the delivery area.
    """

    def __init__(self, context: DeliveryContextIF, underlying: Destination):
        super().__init__(underlying)
        self.context = context

    def add_file(self, file: pathlib.Path, relative_path: str):
        self.underlying.add_file(file, relative_path)

    def add_directory(self, directory: pathlib.Path, relative_path: str):
        self.underlying.add_directory(directory, relative_path)

    def close(self):
        self.underlying.close()
        # I beleive this gives us the request id, but I may be wrong
        # I also assume that request id is what we want, may not be true
        tar_name = os.path.basename(os.path.normpath(self.context.source))
        shutil.make_archive(self.underlying.path + "/" + tar_name, "gztar", self.underlying.path)
        # clear out the directory, leaving the tar
        for filename in os.listdir(self.underlying.path):
            file_path = os.path.join(self.underlying.path, filename)
            try:
                if filename != tar_name + ".tar.gz":
                    if os.path.isfile(file_path) or os.path.islink(file_path):
                        os.unlink(file_path)
                    elif os.path.isdir(file_path):
                        shutil.rmtree(file_path)
            except Exception as e:
                raise e
                # print('Failed to delete %s. Reason: %s' % (file_path, e))


# class ChecksumDecorator(DestinationDecorator):
#     """
#     This decorator ensures that an MD5SUM file appears in the underlying
#     destination after all the files are added, and the contents of that
#     file are, as one would expect, the MD5SUMs of the files added to the
#     destination, in the format that ``md5sum -c`` expects.
#     """
#
#     def __init__(self, context: "DeliveryContext", underlying: Destination):
#         super().__init__(underlying)
#         self.context = context
#
#     def add_file(self, file: pathlib.Path, relative_path: str):
#         self.underlying.add_file(file, relative_path)
#
#     def add_directory(self, directory: pathlib.Path, relative_path: str):
#         self.underlying.add_directory(directory, relative_path)
#
#     def close(self):
#         self.underlying.close()
#         hasher = hashlib.md5()
#         hasher = (fname, self.hash_bytestr_iter(self.file_as_blockiter(open(fname, 'rb')), hasher)) /
#         for fname in glob.iglob(self.underlying.path + '**/**', recursive=True)
#         checksum_file = open(self.underlying.path / "MD5SUMS", "w")
#         checksum_file.write(hasher.hexdigest())
#         checksum_file.close()
#
##   the following is reported to be a memory efficient way to get md5 checksum
##   https://stackoverflow.com/questions/3431825/generating-an-md5-checksum-of-a-file > post by Omnifarious
#    def hash_bytestr_iter(bytesiter, hasher):
#        for block in bytesiter:
#            hasher.update(block)
#        return hasher
#
#    def file_as_blockiter(afile, blocksize=65536):
#        with afile:
#            block = afile.read(blocksize)
#            while len(block) > 0:
#                yield block
#                block = afile.read(blocksize)

# class SubdirectoryDecorator(DestinationDecorator):
#     def __init__(self, underlying: Destination, subdirectory: str):
#         super().__init__(underlying)
#         self.subdirectory = subdirectory
#
#     def add_file(self, relative_path: str, file: pathlib.Path):
#         self.underlying.add_file(self.subdirectory + "/" + relative_path, file)


class LocalDestination(Destination):
    """
    LocalDestination is for delivering to a local directory on the filesystem.
    """

    def add_file(self, file: pathlib.Path, relative_path: str):
        """
        Copy contents of file to new file with path relative_path; creates directories if they don't exis

        :param file: Source file whose contents are getting copied
        :param relative_path: Relative path to new file in destination
        """
        try:
            # if don't care about file metadata we could use copy()
            shutil.copy2(file, relative_path)
        except shutil.Error as err:
            raise err

    def add_directory(self, directory: pathlib.Path, relative_path: str):
        """
        Add each file in a directory to destination directory; creates directories if they don't exist

        :param directory: Source directory whose files are getting copied
        :param relative_path: Relative path to new directory location in destination
        """
        try:
            shutil.copytree(directory, str(self.path) + "/" + relative_path, dirs_exist_ok=True)
        except shutil.Error as err:
            raise err

    def close(self):
        """
        Nothing special actually needs to be done for local deliveries
        when we close the destination.
        """
        pass


class DestinationBuilder:
    """
    To facilitate building a stack of destination and its decorators.
    """

    def __init__(self, context: DeliveryContextIF):
        self.context = context
        self._destination = None

    def local(self, path: pathlib.Path):
        """Add a local destination with the given path"""
        self._destination = LocalDestination(self.context, path)
        return self

    def tar(self):
        """Add the tar decorator to the destination"""
        self._destination = TarDecorator(self.context, self._destination)
        return self

    def build(self):
        """Create the destination"""
        # We *always* want checksums, so that goes on last here
        # return ChecksumDecorator(self.context, self._destination)
        return self._destination
